-- ********************************************************************************
-- Binarizer for HEVC H.265       
-- 01/03/2017                 
-- Norwegian University of Science and Technology              
-- Lars Erik Songe Paulsen             
-- ********************************************************************************

-- ********************************************************************************
-- TODO LIST:    
-- ********************************************************************************
-- 
-- ********************************************************************************

--pragma synthesis_off
  -- your simulation-only code
--pragma synthesis_on

library ieee;
library std;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all; -- http://www1.pldworld.com/@xilinx/html/technote/tool/manual/15i_doc/fndtn/vhd/vhd10_3.htm

library work;
use work.BinarizerParameters.all;

entity Binarizer is

    port    
    (
    Clk                         : in std_logic;
    DataIn                      : in std_logic_vector((16*CoeffWidth)-1 downto 0);
    ScanDir                     : in std_logic_vector(1 downto 0);
    Resetn                      : in std_logic;
    StartBinarizer              : in std_logic;
    DataOut                     : out std_logic_vector(OutputWidth-1 downto 0); 
    DataLength                  : out std_logic_vector(OutputWidthLength-1 downto 0);
    PrefixLength                : out std_logic_vector(OutputWidthLength-1 downto 0);
    Finished                    : out std_logic
    );

end Binarizer;

architecture struct of Binarizer is

    -- ---------------------------------------------------------------------------
    -- Signal declarations
    -- ---------------------------------------------------------------------------

    type BinarizeStateType is(
                    read_input, 
                    write_last_sig_coeff_x_prefix, 
                    write_last_sig_coeff_y_prefix, 
                    write_sig_coeff_flag,
                    write_coeff_abs_level_greater1,
                    write_coeff_abs_level_greater2,
                    sync,
                    write_coeff_abs_level_remaining,
                    write_coeff_sign_flag,
                    write_finished
                );

    type transform_block is array(15 downto 0) of integer range -(2**(coeffwidth-1)) to (2**(coeffwidth-1));

    signal BinarizeState : BinarizeStateType;
    
    signal coefficients : Transform_Block;
    signal coeff_abs_level_greater1_flag : std_logic_vector(0 to 15);
    signal coeff_abs_level_greater2_flag : std_logic_vector(0 to 15);
    signal sig_coeff_flag : std_logic_vector(0 to 15);
    signal ABS_index : integer range 0 to 15;
    signal ABS_started : std_logic;
    signal ABS_done : std_logic;
    signal ABS_level_writeout : integer range 0 to 32767;
    signal k : integer range 0 to 4;
    
begin

    Next_ABS : process(Clk, DataIn, Resetn, StartBinarizer)
        variable ABS_index_cycler : integer range 0 to 15;
        
    begin
        if Resetn = '0' then
            ABS_index_cycler := 0;
            ABS_index <= 0;
            ABS_done  <= '0';
        elsif rising_edge(Clk) then
            case BinarizeState is                               
                when sync | write_coeff_abs_level_remaining =>
                    find_next_ALG2_loop : for ABS_index_cycler in 0 to 15 loop
                        if (ABS_index_cycler <= ABS_index) then --<<<<<<<<< Here be issues
                        --spin TODO FIX end condition
                        else
                            if(coeff_abs_level_greater1_flag(ABS_index_cycler) = '1') then
                                report "ALG found at ABS_index[" & integer'image(ABS_index_cycler) & "]";
                                ABS_index <= ABS_index_cycler;
                                ABS_level_writeout <= coefficients(ABS_index_cycler);
                                if(ABS_index_cycler = 15) then
                                    ABS_done <= '1';
                                else
                                    exit find_next_ALG2_loop;
                                end if;
                            end if;
                        end if;
                    end loop;
                when others =>
                    ABS_index <= 0;
                    ABS_done  <= '0';
            end case;
        end if;
    end process;

    Binarize : process(Clk, DataIn, Resetn, StartBinarizer)
    
    -- ---------------------------------------------------------------------------
    -- Variable declarations
    -- ---------------------------------------------------------------------------
    
    variable sig_coeff_flag_found : std_logic;
    variable sig_coeff_flag_index : integer range OutputWidth-1 downto OutputWidth-14;
    variable coeff_sign_flag : std_logic_vector(0 to 15);
    variable ALG1_index  : integer range 0 to 8;   
    variable SIGN_index  : integer range 0 to 15;   
    variable Scan_Direction : Scan_Directions := DiagonalScan; -- TODO: Add all scan directions
    
    begin
        if Resetn = '0' then
            sig_coeff_flag <= (others => '0');
            sig_coeff_flag_found := '0';
            sig_coeff_flag_index := OutputWidth-1;
            ALG1_index     := 0;
            SIGN_index     := 0;
            DataOut        <= (others => '0');
            DataLength     <= (others => '0');
            BinarizeState  <= read_input;
            Finished       <= '1';
            coeff_abs_level_greater1_flag <= (others => '0');
            coeff_abs_level_greater2_flag <= (others => '0');
            k   <= 0;
            PrefixLength <= (others => '0');
            
        elsif rising_edge(Clk) then
            case BinarizeState is
                when read_input =>
                    BinarizeState <= write_last_sig_coeff_x_prefix;
                    Finished      <= '0';
                    
                    for i in 0 to 15 loop
                        -- ---------------------------------------------------------------------------
                        -- Read absolute values.
                        -- ---------------------------------------------------------------------------  
                        coefficients(Scan_Direction(15-i))  <= to_integer(abs(signed(DataIn(((CoeffWidth*(i+1))-1)   downto  (CoeffWidth*i)))));
                        
                        -- ---------------------------------------------------------------------------
                        -- Read sign.
                        -- ---------------------------------------------------------------------------
                        coeff_sign_flag(Scan_Direction(15-i))   := DataIn((CoeffWidth*(i+1))-1);
                        
                        -- ---------------------------------------------------------------------------
                        -- Find non-zero data.
                        -- ---------------------------------------------------------------------------  
                        sig_coeff_flag(Scan_Direction(15-i))  <= or_reduce(DataIn(((CoeffWidth*(i+1))-2)   downto  (CoeffWidth*i))); 
                    
                        -- ---------------------------------------------------------------------------
                        -- Find >1 data and >2 data. Procedure dependent on sign bit.
                        -- ---------------------------------------------------------------------------  
                        case DataIn((CoeffWidth*(i+1))-1) is
                            when '0' =>
                                coeff_abs_level_greater1_flag(Scan_Direction(15-i))    <= or_reduce(DataIn(((CoeffWidth*(i+1))-2)  downto  ((CoeffWidth*i)+1)));
                                coeff_abs_level_greater2_flag(Scan_Direction(15-i))    <= (DataIn(CoeffWidth*i)   and DataIn((CoeffWidth*i)+1)) 
                                                                                       or or_reduce(DataIn(((CoeffWidth*(i+1))-2) downto ((CoeffWidth*i)+2))); 
                            when '1' =>
                                coeff_abs_level_greater1_flag(Scan_Direction(15-i))    <= nand_reduce(DataIn(((CoeffWidth*(i+1))-2)  downto  (CoeffWidth*i)));
                                coeff_abs_level_greater2_flag(Scan_Direction(15-i))    <= (DataIn(CoeffWidth*i)   and (not(DataIn((CoeffWidth*i)+1)))) 
                                                                                       or nand_reduce(DataIn(((CoeffWidth*(i+1))-2)  downto((CoeffWidth*i)+2))); 
                            when others =>
                        end case;
                    end loop;

                when write_last_sig_coeff_x_prefix =>

                        for i in 0 to 15 loop
                            report "coefficients(" & integer'image(i) & "): " & integer'image(coefficients(i));
                        end loop;

                    report "sig_coeff_flag: " & integer'image(conv_integer(sig_coeff_flag));
                    report "coeff_abs_level_greater1_flag: " & integer'image(conv_integer(coeff_abs_level_greater1_flag));
                    report "coeff_abs_level_greater2_flag: " & integer'image(conv_integer(coeff_abs_level_greater2_flag));
                    report "coeff_sign_flag: " & integer'image(conv_integer(coeff_sign_flag));

                    BinarizeState <= write_last_sig_coeff_y_prefix;
                    write_last_sig_coeff_x_prefix_loop : for index in 0 to 15 loop
                        if(sig_coeff_flag(index) = '1') then
                            case index is                                
                                when 0  =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "111"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength)); 
                                when 1  =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "111"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength)); 
                                when 2  =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "110"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength)); 
                                when 3  =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "111"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength)); 
                                when 4  =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "110"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength)); 
                                when 5  =>  DataOut(OutputWidth-1 downto OutputWidth-2) <= "10";  
                                            DataLength <= std_logic_vector(to_unsigned(2,OutputWidthLength)); 
                                when 6  =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "111"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength)); 
                                when 7  =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "110"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength));
                                when 8  =>  DataOut(OutputWidth-1 downto OutputWidth-2) <= "10";  
                                            DataLength <= std_logic_vector(to_unsigned(2,OutputWidthLength));
                                when 9  =>  DataOut(OutputWidth-1)                      <= '0';   
                                            DataLength <= std_logic_vector(to_unsigned(1,OutputWidthLength));
                                when 10 =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "110"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength));
                                when 11 =>  DataOut(OutputWidth-1 downto OutputWidth-2) <= "10";  
                                            DataLength <= std_logic_vector(to_unsigned(2,OutputWidthLength));
                                when 12 =>  DataOut(OutputWidth-1)                      <= '0';   
                                            DataLength <= std_logic_vector(to_unsigned(1,OutputWidthLength));
                                when 13 =>  DataOut(OutputWidth-1 downto OutputWidth-2) <= "10";  
                                            DataLength <= std_logic_vector(to_unsigned(2,OutputWidthLength));
                                when 14 =>  DataOut(OutputWidth-1)                      <= '0';   
                                            DataLength <= std_logic_vector(to_unsigned(1,OutputWidthLength));
                                when 15 =>  DataOut(OutputWidth-1)                      <= '0';   
                                            DataLength <= std_logic_vector(to_unsigned(1,OutputWidthLength));
                            end case;
                            exit write_last_sig_coeff_x_prefix_loop;
                        end if;
                    end loop;
                when write_last_sig_coeff_y_prefix =>
                    BinarizeState <= write_sig_coeff_flag;
                    write_last_sig_coeff_y_prefix_loop : for index in 0 to 15 loop
                        if(sig_coeff_flag(index) = '1') then
                            case index is                                
                                when 0  =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "111"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength)); 
                                when 1  =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "110"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength)); 
                                when 2  =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "111"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength)); 
                                when 3  =>  DataOut(OutputWidth-1 downto OutputWidth-2) <= "10";  
                                            DataLength <= std_logic_vector(to_unsigned(2,OutputWidthLength)); 
                                when 4  =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "110"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength)); 
                                when 5  =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "110"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength)); 
                                when 6  =>  DataOut(OutputWidth-1)                      <= '0';   
                                            DataLength <= std_logic_vector(to_unsigned(1,OutputWidthLength)); 
                                when 7  =>  DataOut(OutputWidth-1 downto OutputWidth-2) <= "10";  
                                            DataLength <= std_logic_vector(to_unsigned(2,OutputWidthLength));
                                when 8  =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "110"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength));
                                when 9  =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "111"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength));
                                when 10 =>  DataOut(OutputWidth-1)                      <= '0';   
                                            DataLength <= std_logic_vector(to_unsigned(1,OutputWidthLength));
                                when 11 =>  DataOut(OutputWidth-1 downto OutputWidth-2) <= "10";  
                                            DataLength <= std_logic_vector(to_unsigned(2,OutputWidthLength));
                                when 12 =>  DataOut(OutputWidth-1 downto OutputWidth-3) <= "110"; 
                                            DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength));
                                when 13 =>  DataOut(OutputWidth-1)                      <= '0';   
                                            DataLength <= std_logic_vector(to_unsigned(1,OutputWidthLength));
                                when 14 =>  DataOut(OutputWidth-1 downto OutputWidth-2) <= "10";  
                                            DataLength <= std_logic_vector(to_unsigned(2,OutputWidthLength));
                                when 15 =>  DataOut(OutputWidth-1)                      <= '0';   
                                            DataLength <= std_logic_vector(to_unsigned(1,OutputWidthLength));
                            end case;
                            exit write_last_sig_coeff_y_prefix_loop;
                        end if;
                    end loop;
                when write_sig_coeff_flag =>
                    BinarizeState <= write_coeff_abs_level_greater1;
                    write_sig_coeff_flag_loop : for index in 0 to 15 loop
                        if sig_coeff_flag_found = '1' then
                            DataOut(sig_coeff_flag_index) <= sig_coeff_flag(index);
                            sig_coeff_flag_index := sig_coeff_flag_index - 1; 
                            if(sig_coeff_flag(index) = '1') then
                                coefficients(index) <= coefficients(index) - 1;
                            end if;
                        elsif(sig_coeff_flag(index) = '1') then
                            sig_coeff_flag_found := '1';
                            DataLength <= std_logic_vector(to_unsigned(15-index,OutputWidthLength));
                            coefficients(index) <= coefficients(index) - 1;                         
                        end if;
                    end loop;   

                when write_coeff_abs_level_greater1 =>
                    BinarizeState <= write_coeff_abs_level_greater2;
                    write_coeff_abs_level_greater_1_loop : for index in 0 to 15 loop
                        if sig_coeff_flag(index) = '1' and ALG1_index < 8 then
                            if( coeff_abs_level_greater1_flag(index) = '1') then
                                DataOut(OutputWidth-1-ALG1_index) <= '1';
                                report "ALG1:coefficients(" & integer'image(index) & "): " & integer'image(coefficients(index));
                                coefficients(index) <= coefficients(index) - 1;
                                report "ALG1:coefficients(" & integer'image(index) & "): " & integer'image(coefficients(index));
                            else
                                DataOut(OutputWidth-1-ALG1_index) <= '0';
                            end if;
                            ALG1_index := ALG1_index + 1;
                        end if;   
                    end loop; 

                    DataLength <= std_logic_vector(to_unsigned(ALG1_index,OutputWidthLength));

                when write_coeff_abs_level_greater2 =>
                    BinarizeState <= sync;
                    write_coeff_abs_level_greater_2_loop : for index in 0 to 15 loop
                        if sig_coeff_flag(index) = '1' then                            
                            if (coeff_abs_level_greater2_flag(index) = '1') then
                                DataOut(OutputWidth-1) <= '1';
                                report "ALG2:coefficients(" & integer'image(index) & "): " & integer'image(coefficients(index));
                                coefficients(index) <= coefficients(index) - 1;
                                report "ALG2:coefficients(" & integer'image(index) & "): " & integer'image(coefficients(index));
                            else
                                DataOut(OutputWidth-1) <= '0';
                            end if;
                            DataLength <= std_logic_vector(to_unsigned(1,OutputWidthLength));
                            exit write_coeff_abs_level_greater_2_loop;
                        end if;   
                    end loop; 
                    
                when sync =>
                    BinarizeState <= write_coeff_abs_level_remaining;

                        for i in 0 to 15 loop
                            report "coefficients(" & integer'image(i) & "): " & integer'image(coefficients(i));
                        end loop;

                when write_coeff_abs_level_remaining =>
                    report "write_coeff_abs_level_remaining: @ k: " & integer'image(k) & " ABS_level_writeout: " & integer'image(ABS_level_writeout);
                    if(ABS_done = '1') then
                        BinarizeState <= write_coeff_sign_flag;
                    end if;
                    case k is
                        when 0 =>
                            case (ABS_level_writeout) is
                                when 0 =>
                                    DataLength <= std_logic_vector(to_unsigned(0,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(1,OutputWidthLength));
                                when 1 =>
                                    DataLength <= std_logic_vector(to_unsigned(0,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(2,OutputWidthLength));
                                when 2 =>
                                    DataLength <= std_logic_vector(to_unsigned(0,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(3,OutputWidthLength));
                                when 3 =>
                                    DataLength <= std_logic_vector(to_unsigned(0,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(4,OutputWidthLength));
                                when 4 to 5 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-1) <= std_logic_vector(to_unsigned(ABS_level_writeout-4,1));
                                    DataLength <= std_logic_vector(to_unsigned(1,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(5,OutputWidthLength));
                                    k <= 1;
                                when 6 to 9 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-2) <= std_logic_vector(to_unsigned(ABS_level_writeout-6,2));
                                    DataLength <= std_logic_vector(to_unsigned(2,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(6,OutputWidthLength));
                                when 10 to 17 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-3) <= std_logic_vector(to_unsigned(ABS_level_writeout-10,3));
                                    DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(7,OutputWidthLength));
                                when 18 to 33 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-4) <= std_logic_vector(to_unsigned(ABS_level_writeout-18,4));
                                    DataLength <= std_logic_vector(to_unsigned(4,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(8,OutputWidthLength));
                                when 34 to 65 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-5) <= std_logic_vector(to_unsigned(ABS_level_writeout-34,5));
                                    DataLength <= std_logic_vector(to_unsigned(5,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(9,OutputWidthLength));
                                when 66 to 129 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-6) <= std_logic_vector(to_unsigned(ABS_level_writeout-66,6));
                                    DataLength <= std_logic_vector(to_unsigned(6,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(10,OutputWidthLength));
                                when 130 to 257 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-7) <= std_logic_vector(to_unsigned(ABS_level_writeout-130,7));
                                    DataLength <= std_logic_vector(to_unsigned(7,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(11,OutputWidthLength));
                                when 258 to 513 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-8) <= std_logic_vector(to_unsigned(ABS_level_writeout-258,8));
                                    DataLength <= std_logic_vector(to_unsigned(8,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(12,OutputWidthLength));
                                when 514 to 1025 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-9) <= std_logic_vector(to_unsigned(ABS_level_writeout-514,9));
                                    DataLength <= std_logic_vector(to_unsigned(9,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(13,OutputWidthLength));
                                when 1026 to 2049 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-10) <= std_logic_vector(to_unsigned(ABS_level_writeout-1026,10));
                                    DataLength <= std_logic_vector(to_unsigned(10,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(14,OutputWidthLength));
                                when 2050 to 4097 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-11) <= std_logic_vector(to_unsigned(ABS_level_writeout-2050,11));
                                    DataLength <= std_logic_vector(to_unsigned(11,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(15,OutputWidthLength));
                                when 4098 to 8193 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-12) <= std_logic_vector(to_unsigned(ABS_level_writeout-4098,12));
                                    DataLength <= std_logic_vector(to_unsigned(12,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(16,OutputWidthLength));
                                when 8194 to 16385 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-13) <= std_logic_vector(to_unsigned(ABS_level_writeout-8194,13));
                                    DataLength <= std_logic_vector(to_unsigned(13,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(17,OutputWidthLength));
                                when 16386 to 32767 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-14) <= std_logic_vector(to_unsigned(ABS_level_writeout-16386,14));
                                    DataLength <= std_logic_vector(to_unsigned(14,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(18,OutputWidthLength));
                                when others     =>
                            end case;
                        when 1 =>
                            case (ABS_level_writeout) is
                                when 0 to 1 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-1) <= std_logic_vector(to_unsigned(ABS_level_writeout,1));
                                    DataLength <= std_logic_vector(to_unsigned(1,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(1,OutputWidthLength));
                                when 2 to 3 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-1) <= std_logic_vector(to_unsigned(ABS_level_writeout-2,1));
                                    DataLength <= std_logic_vector(to_unsigned(1,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(2,OutputWidthLength));
                                when 4 to 5 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-1) <= std_logic_vector(to_unsigned(ABS_level_writeout-4,1));
                                    DataLength <= std_logic_vector(to_unsigned(1,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(3,OutputWidthLength));
                                when 6 to 7 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-1) <= std_logic_vector(to_unsigned(ABS_level_writeout-6,1));
                                    DataLength <= std_logic_vector(to_unsigned(1,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(4,OutputWidthLength));
                                when 8 to 11 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-2) <= std_logic_vector(to_unsigned(ABS_level_writeout-8,2));
                                    DataLength <= std_logic_vector(to_unsigned(2,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(5,OutputWidthLength));
                                when 12 to 19 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-3) <= std_logic_vector(to_unsigned(ABS_level_writeout-12,3));
                                    DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(6,OutputWidthLength));
                                when 20 to 35 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-4) <= std_logic_vector(to_unsigned(ABS_level_writeout-20,4));
                                    DataLength <= std_logic_vector(to_unsigned(4,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(7,OutputWidthLength));
                                when 36 to 67 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-5) <= std_logic_vector(to_unsigned(ABS_level_writeout-36,5));
                                    DataLength <= std_logic_vector(to_unsigned(5,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(8,OutputWidthLength));
                                when 68 to 131 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-6) <= std_logic_vector(to_unsigned(ABS_level_writeout-68,6));
                                    DataLength <= std_logic_vector(to_unsigned(6,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(9,OutputWidthLength));
                                when 132 to 259 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-7) <= std_logic_vector(to_unsigned(ABS_level_writeout-132,7));
                                    DataLength <= std_logic_vector(to_unsigned(7,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(10,OutputWidthLength));
                                when 260 to 515 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-8) <= std_logic_vector(to_unsigned(ABS_level_writeout-260,8));
                                    DataLength <= std_logic_vector(to_unsigned(8,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(11,OutputWidthLength));
                                when 516 to 1027 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-9) <= std_logic_vector(to_unsigned(ABS_level_writeout-516,9));
                                    DataLength <= std_logic_vector(to_unsigned(9,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(12,OutputWidthLength));
                                when 1028 to 2051 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-10) <= std_logic_vector(to_unsigned(ABS_level_writeout-1028,10));
                                    DataLength <= std_logic_vector(to_unsigned(10,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(13,OutputWidthLength));
                                when 2052 to 4099 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-11) <= std_logic_vector(to_unsigned(ABS_level_writeout-2052,11));
                                    DataLength <= std_logic_vector(to_unsigned(11,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(14,OutputWidthLength));
                                when 4100 to 8195 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-12) <= std_logic_vector(to_unsigned(ABS_level_writeout-4100,12));
                                    DataLength <= std_logic_vector(to_unsigned(12,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(15,OutputWidthLength));
                                when 8196 to 16387 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-13) <= std_logic_vector(to_unsigned(ABS_level_writeout-8196,13));
                                    DataLength <= std_logic_vector(to_unsigned(13,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(16,OutputWidthLength));
                                when 16388 to 32767 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-14) <= std_logic_vector(to_unsigned(ABS_level_writeout-16388,14));
                                    DataLength <= std_logic_vector(to_unsigned(14,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(17,OutputWidthLength));
                                when others     =>
                            end case;
                        when 2 =>
                            case (ABS_level_writeout) is
                                when 0 to 3 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-2) <= std_logic_vector(to_unsigned(ABS_level_writeout,2));
                                    DataLength <= std_logic_vector(to_unsigned(2,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(1,OutputWidthLength));
                                when 4 to 7 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-2) <= std_logic_vector(to_unsigned(ABS_level_writeout-4,2));
                                    DataLength <= std_logic_vector(to_unsigned(2,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(2,OutputWidthLength));
                                when 8 to 11 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-2) <= std_logic_vector(to_unsigned(ABS_level_writeout-8,2));
                                    DataLength <= std_logic_vector(to_unsigned(2,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(3,OutputWidthLength));
                                when 12 to 15 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-2) <= std_logic_vector(to_unsigned(ABS_level_writeout-12,2));
                                    DataLength <= std_logic_vector(to_unsigned(2,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(4,OutputWidthLength));
                                when 16 to 23 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-3) <= std_logic_vector(to_unsigned(ABS_level_writeout-16,3));
                                    DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(5,OutputWidthLength));
                                when 24 to 39 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-4) <= std_logic_vector(to_unsigned(ABS_level_writeout-24,4));
                                    DataLength <= std_logic_vector(to_unsigned(4,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(6,OutputWidthLength));
                                when 40 to 71 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-5) <= std_logic_vector(to_unsigned(ABS_level_writeout-40,5));
                                    DataLength <= std_logic_vector(to_unsigned(5,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(7,OutputWidthLength));
                                when 72 to 135 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-6) <= std_logic_vector(to_unsigned(ABS_level_writeout-72,6));
                                    DataLength <= std_logic_vector(to_unsigned(6,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(8,OutputWidthLength));
                                when 136 to 263 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-7) <= std_logic_vector(to_unsigned(ABS_level_writeout-136,7));
                                    DataLength <= std_logic_vector(to_unsigned(7,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(9,OutputWidthLength));
                                when 264 to 519 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-8) <= std_logic_vector(to_unsigned(ABS_level_writeout-264,8));
                                    DataLength <= std_logic_vector(to_unsigned(8,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(10,OutputWidthLength));
                                when 520 to 1031 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-9) <= std_logic_vector(to_unsigned(ABS_level_writeout-520,9));
                                    DataLength <= std_logic_vector(to_unsigned(9,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(11,OutputWidthLength));
                                when 1032 to 2055 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-10) <= std_logic_vector(to_unsigned(ABS_level_writeout-1032,10));
                                    DataLength <= std_logic_vector(to_unsigned(10,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(12,OutputWidthLength));
                                when 2056 to 4103 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-11) <= std_logic_vector(to_unsigned(ABS_level_writeout-2056,11));
                                    DataLength <= std_logic_vector(to_unsigned(11,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(13,OutputWidthLength));
                                when 4104 to 8199 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-12) <= std_logic_vector(to_unsigned(ABS_level_writeout-4104,12));
                                    DataLength <= std_logic_vector(to_unsigned(12,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(14,OutputWidthLength));
                                when 8200 to 16391 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-13) <= std_logic_vector(to_unsigned(ABS_level_writeout-8200,13));
                                    DataLength <= std_logic_vector(to_unsigned(13,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(15,OutputWidthLength));
                                when 16392 to 32767 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-14) <= std_logic_vector(to_unsigned(ABS_level_writeout-16392,14));
                                    DataLength <= std_logic_vector(to_unsigned(14,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(16,OutputWidthLength));
                                when others     =>
                            end case;
                        when 3 =>
                            case (ABS_level_writeout) is
                                when 0 to 7 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-3) <= std_logic_vector(to_unsigned(ABS_level_writeout,3));
                                    DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(1,OutputWidthLength));
                                when 8 to 15 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-3) <= std_logic_vector(to_unsigned(ABS_level_writeout-8,3));
                                    DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(2,OutputWidthLength));
                                when 16 to 23 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-3) <= std_logic_vector(to_unsigned(ABS_level_writeout-16,3));
                                    DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(3,OutputWidthLength));
                                when 24 to 31 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-3) <= std_logic_vector(to_unsigned(ABS_level_writeout-24,3));
                                    DataLength <= std_logic_vector(to_unsigned(3,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(4,OutputWidthLength));
                                when 32 to 47 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-4) <= std_logic_vector(to_unsigned(ABS_level_writeout-32,4));
                                    DataLength <= std_logic_vector(to_unsigned(4,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(5,OutputWidthLength));
                                when 48 to 79 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-5) <= std_logic_vector(to_unsigned(ABS_level_writeout-48,5));
                                    DataLength <= std_logic_vector(to_unsigned(5,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(6,OutputWidthLength));
                                when 80 to 143 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-6) <= std_logic_vector(to_unsigned(ABS_level_writeout-80,6));
                                    DataLength <= std_logic_vector(to_unsigned(6,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(7,OutputWidthLength));
                                when 144 to 271 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-7) <= std_logic_vector(to_unsigned(ABS_level_writeout-144,7));
                                    DataLength <= std_logic_vector(to_unsigned(7,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(8,OutputWidthLength));
                                when 272 to 527 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-8) <= std_logic_vector(to_unsigned(ABS_level_writeout-272,8));
                                    DataLength <= std_logic_vector(to_unsigned(8,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(9,OutputWidthLength));
                                when 528 to 1039 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-9) <= std_logic_vector(to_unsigned(ABS_level_writeout-528,9));
                                    DataLength <= std_logic_vector(to_unsigned(9,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(10,OutputWidthLength));
                                when 1040 to 2063 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-10) <= std_logic_vector(to_unsigned(ABS_level_writeout-1040,10));
                                    DataLength <= std_logic_vector(to_unsigned(10,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(11,OutputWidthLength));
                                when 2064 to 4111 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-11) <= std_logic_vector(to_unsigned(ABS_level_writeout-2064,11));
                                    DataLength <= std_logic_vector(to_unsigned(11,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(12,OutputWidthLength));
                                when 4112 to 8207 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-12) <= std_logic_vector(to_unsigned(ABS_level_writeout-4112,12));
                                    DataLength <= std_logic_vector(to_unsigned(12,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(13,OutputWidthLength));
                                when 8208 to 16399 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-13) <= std_logic_vector(to_unsigned(ABS_level_writeout-8208,13));
                                    DataLength <= std_logic_vector(to_unsigned(13,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(14,OutputWidthLength));
                                when 16400 to 32767 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-14) <= std_logic_vector(to_unsigned(ABS_level_writeout-16400,14));
                                    DataLength <= std_logic_vector(to_unsigned(14,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(15,OutputWidthLength));
                                when others     =>
                            end case;
                        when 4 =>
                            case (ABS_level_writeout) is
                                when 0 to 15 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-4) <= std_logic_vector(to_unsigned(ABS_level_writeout,4));
                                    DataLength <= std_logic_vector(to_unsigned(4,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(1,OutputWidthLength));
                                when 16 to 31 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-4) <= std_logic_vector(to_unsigned(ABS_level_writeout-16,4));
                                    DataLength <= std_logic_vector(to_unsigned(4,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(2,OutputWidthLength));
                                when 32 to 47 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-4) <= std_logic_vector(to_unsigned(ABS_level_writeout-32,4));
                                    DataLength <= std_logic_vector(to_unsigned(4,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(3,OutputWidthLength));
                                when 48 to 63 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-4) <= std_logic_vector(to_unsigned(ABS_level_writeout-48,4));
                                    DataLength <= std_logic_vector(to_unsigned(4,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(4,OutputWidthLength));
                                when 64 to 95 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-5) <= std_logic_vector(to_unsigned(ABS_level_writeout-64,5));
                                    DataLength <= std_logic_vector(to_unsigned(5,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(5,OutputWidthLength));
                                when 96 to 159 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-6) <= std_logic_vector(to_unsigned(ABS_level_writeout-96,6));
                                    DataLength <= std_logic_vector(to_unsigned(6,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(6,OutputWidthLength));
                                when 160 to 287 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-7) <= std_logic_vector(to_unsigned(ABS_level_writeout-160,7));
                                    DataLength <= std_logic_vector(to_unsigned(7,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(7,OutputWidthLength));
                                when 288 to 543 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-8) <= std_logic_vector(to_unsigned(ABS_level_writeout-288,8));
                                    DataLength <= std_logic_vector(to_unsigned(8,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(8,OutputWidthLength));
                                when 544 to 1055 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-9) <= std_logic_vector(to_unsigned(ABS_level_writeout-544,9));
                                    DataLength <= std_logic_vector(to_unsigned(9,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(9,OutputWidthLength));
                                when 1056 to 2079 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-10) <= std_logic_vector(to_unsigned(ABS_level_writeout-1056,10));
                                    DataLength <= std_logic_vector(to_unsigned(10,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(10,OutputWidthLength));
                                when 2080 to 4127 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-11) <= std_logic_vector(to_unsigned(ABS_level_writeout-2080,11));
                                    DataLength <= std_logic_vector(to_unsigned(11,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(11,OutputWidthLength));
                                when 4128 to 8223 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-12) <= std_logic_vector(to_unsigned(ABS_level_writeout-4128,12));
                                    DataLength <= std_logic_vector(to_unsigned(12,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(12,OutputWidthLength));
                                when 8224 to 16415 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-13) <= std_logic_vector(to_unsigned(ABS_level_writeout-8224,13));
                                    DataLength <= std_logic_vector(to_unsigned(13,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(13,OutputWidthLength));
                                when 16416 to 32767 =>
                                    DataOut(OutputWidth-1 downto OutputWidth-14) <= std_logic_vector(to_unsigned(ABS_level_writeout-16416,14));
                                    DataLength <= std_logic_vector(to_unsigned(14,OutputWidthLength));
                                    PrefixLength <= std_logic_vector(to_unsigned(14,OutputWidthLength));
                                when others     =>
                            end case;
                    end case;

                    if(k<4 and (ABS_level_writeout) > (3*(2**k))) then
                        k <= k+1;
                    end if;

                when write_coeff_sign_flag =>
                    BinarizeState <= write_finished;
                    write_coeff_sign_flag_loop : for index in 0 to 15 loop
                        if sig_coeff_flag(index) = '1' then
                            if coeff_sign_flag(index) = '1' then
                                DataOut(OutputWidth-1 - SIGN_index) <= '1';
                            else
                                DataOut(OutputWidth-1 - SIGN_index) <= '0';
                            end if;
                            SIGN_index := SIGN_index + 1;
                        end if;
                        DataLength <= std_logic_vector(to_unsigned(SIGN_index,OutputWidthLength));
                    end loop;
                when others =>
                    BinarizeState <= write_finished;
                    Finished <= '1';
                    DataOut        <= (others => '0');
                    DataLength     <= (others => '0');
            end case;
        end if;
    end process;
end struct;