library ieee;
library std;

use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_unsigned.all;
use ieee.math_real.all;

package BinarizerParameters is
    
    type Scan_Directions is array (0 to 15) of integer range 0 to 15;

    -- Read sequence for 4x4 TB: (0,0), (0,1), (0,2), (0,3), 
    --                           (1,0), (1,1), (1,2), (1,3), 
    --                           (2,0), (2,1), (2,2), (2,3), 
    --                           (3,0), (3,1), (3,2), (3,3)
    
    constant VerticalScan   : Scan_Directions := (15,14,13,12,11,10,9,8,7,6,5,4,3,2,1,0);
    constant HorizontalScan : Scan_Directions := (15,11,7,3,14,10,6,2,13,9,5,1,12,8,4,0);
    constant DiagonalScan   : Scan_Directions := (15,14,12,9,13,11,8,5,10,7,4,2,6,3,1,0);
    
    constant CoeffWidth : integer := 16; -- Minimum 8, maximum 16.
    constant OutputWidth : integer := 17; --TODO: Check if ranges is ok.
    constant OutputWidthLength : integer := integer(floor(log2(real(OutputWidth))+real(1))); -- Number of bits required to represent OutputWidth.
    
end BinarizerParameters;