-- ********************************************************************************
-- Bypass coder Testbench for HEVC h.265	      
-- 01/03/2017			      
-- Norwegian University of Science and Technology              
-- Lars Erik Songe Paulsen             
-- ********************************************************************************

library ieee;
library std;
use std.textio.all;
use ieee.std_logic_textio.all;

use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_unsigned.all;

library work;
use work.BypassParameters.all;


entity BypassTestbench is

end BypassTestbench;

architecture struct of BypassTestbench is 


	-- ---------------------------------------------------------------------------
	-- Signal declarations
	-- ---------------------------------------------------------------------------
	
	signal Clk			            : std_logic;
	signal DataIn			        : std_logic_vector(InputWidth-1 downto 0);
    signal DataInLength             : std_logic_vector(InputWidthLength-1 downto 0);
	signal ContextModel			    : std_logic_vector(1 downto 0);
	signal Resetn	        		: std_logic;
	signal StartBypass   		    : std_logic;
	signal DataOut                	: std_logic_vector(OutputWidth-1 downto 0);
	signal DataOutLength            : std_logic_vector(OutputWidthLength-1 downto 0);
    signal Termination              : std_logic;
    signal Finished                 : std_logic;

	-- ---------------------------------------------------------------------------
	-- Test input
	-- ---------------------------------------------------------------------------

    -- Include TestFile.txt in Vivado simulation sources.
    file TestFile : TEXT open read_mode is "C:\Users\lepaulse\Documents\GitHub\CABAC\ContextInitializationHEVC_WF\ContextInitializationHEVC_WF\bin\Debug\TestFile.txt";
    -- Target output
    file TestFileOut : TEXT open write_mode is "C:\Users\lepaulse\Documents\GitHub\CABAC\ContextInitializationHEVC_WF\ContextInitializationHEVC_WF\bin\Debug\DecoderInput.txt";
    
	
	-- ---------------------------------------------------------------------------
	-- Functions
	-- ---------------------------------------------------------------------------
	
	-- Parser from TestFile input
	function string_to_binary(inp: string) return std_logic_vector is 
		variable temp: std_logic_vector(inp'length-1 downto 0) := (others => 'X'); 
        --report "input is: " & integer'image(inp'length);
		begin for i in inp'range loop 
			case inp(i) is 
				when '0' => temp(i-1) := '0'; 
				when '1' => temp(i-1) := '1'; 
				when others => temp(i-1) := 'X'; 
			end case; 
		end loop;
		return temp; 
	end function string_to_binary;

        -- Parser to TestFile output
    function binary_to_string(inp: std_logic_vector) return string is 
        variable temp: string(1 to inp'length) := (others => 'X'); 
        --report "input is: " & integer'image(inp'length);
    begin 
        for i in inp'range loop 
            case inp(InputWidth-i-1) is 
                when '0' => temp(i+1) := '0'; 
                when '1' => temp(i+1) := '1'; 
                when others => temp(i+1) := 'X'; 
            end case; 
        end loop;
        return temp; 
    end function binary_to_string;

        -- Parser to TestFile output
    function binary_to_stringW(inp: std_logic_vector) return string is 
        variable temp: string(1 to inp'length) := (others => 'X'); 
        --report "input is: " & integer'image(inp'length);
    begin 
        for i in inp'range loop 
            case inp(OutputWidthLength-i-1) is 
                when '0' => temp(i+1) := '0'; 
                when '1' => temp(i+1) := '1'; 
                when others => temp(i+1) := 'X'; 
            end case; 
        end loop;
        return temp; 
    end function binary_to_stringW;


	-- ---------------------------------------------------------------------------
	-- DUT component
	-- ---------------------------------------------------------------------------

	component Bypass is
		port 
		(
        Clk      		    		: in std_logic;
        DataIn      			    : in std_logic_vector(InputWidth-1 downto 0);
        DataInLength                : in std_logic_vector(InputWidthLength-1 downto 0);
        ContextModel 	    	    : in std_logic_vector(1 downto 0);
        Resetn			        	: in std_logic;
        StartBypass		    	    : in std_logic;
        DataOut                	 	: out std_logic_vector(OutputWidth-1 downto 0); 
        DataOutLength	            : out std_logic_vector(OutputWidthLength-1 downto 0); 
        Termination                 : in std_logic;
        Finished                    : out std_logic
        );
	end component Bypass;

	
begin

	-- ---------------------------------------------------------------------------
	-- DUT port map
	-- ---------------------------------------------------------------------------
	BypassInstance : Bypass
		port map
		(
		Clk      				=> Clk,
		DataIn 			        => DataIn,
        DataInLength            => DataInLength,
		ContextModel			=> ContextModel,
		Resetn				    => Resetn,
		StartBypass			    => StartBypass,
		DataOut		            => DataOut,
		DataOutLength	        => DataOutLength,
        Termination             => Termination,
        Finished                => Finished
		);

	ClkGen : process
	begin
		Clk <= '1';
		wait for 5 ns;
		Clk <= '0';
		wait for 5 ns;
	end process;

	TestGen : process
		
	-- ---------------------------------------------------------------------------
	-- Variable declarations
	-- ---------------------------------------------------------------------------

    --The status of the arithmetic decoding engine is represented by the variables codIRange and codIOffset. In the
    --initialization procedure of the arithmetic decoding process, codIRange is set equal to 510 and codIOffset is set equal to
    --the value returned from read_bits( 9 ) interpreted as a 9 bit binary representation of an unsigned integer with most
    --significant bit written first.



    --From: 9.2.1.4:
    --The bitstream shall not contain data that result in a value of codIOffset being equal to 510 or 511.
    --NOTE – The description of the arithmetic decoding engine in this Specification utilizes 16 bit register precision. However, a
    --minimum register precision of 9 bits is required for storing the values of the variables codIRange and codIOffset after invocation
    --of the arithmetic decoding process (DecodeBin) as specified in subclause 9.2.3.2. The arithmetic decoding process for a binary
    --decision (DecodeDecision) as specified in subclause 9.2.3.2.1 and the decoding process for a binary decision before termination
    --(DecodeTerminate) as specified in subclause 9.2.3.2.4 require a minimum register precision of 9 bits for the variables codIRange
    --and codIOffset. The bypass decoding process for binary decisions (DecodeBypass) as specified in subclause 9.2.3.2.3 requires a
    --minimum register precision of 10 bits for the variable codIOffset and a minimum register precision of 9 bits for the variable
    --codIRange.
    variable TestString : string(InputWidth downto 1);
    variable TestLine : line;

    variable TestStringOut : string(InputWidth downto 1);
    variable TestLineOut : line;

    variable codIOffset : integer range 0 to 1023 := 0;
    variable codIRange : integer range 0 to 511 := 510;
    variable codIRangeVector : std_logic_vector(8 downto 0);
    begin

        --file_open(TestFileOut, "TestFileOut.txt", write_mode);
        -- offset: 2*345+2*115+2-1 = 921
        DataInLength <= "01100"; -- 12
        Termination <= '0';
        readline(TestFile, TestLine);
        read(TestLine, TestString);
        DataIn <= string_to_binary(TestString);
        Resetn <= '0';
        wait for 5 ns;
        Resetn <= '1';
        StartBypass <= '1';
        wait for 10 ns;
        StartBypass <= '0';
        wait until Finished = '1';
        --write(TestLineOut, conv_integer(DataOutLength), left, OutputWidth);
        --writeline(TestFileOut, TestLineOut);
        write(TestLineOut, DataOut, left, conv_integer(DataOutLength));
        writeline(TestFileOut, TestLineOut);
        while not endfile(TestFile) loop
            readline(TestFile, TestLine);
            read(TestLine, TestString);
            DataIn <= string_to_binary(TestString);
            wait for 5 ns;
            StartBypass <= '1';
            wait for 10 ns;
            StartBypass <= '0';
            wait until Finished = '1';
            --write(TestLineOut, conv_integer(DataOutLength), left, OutputWidth);
            --writeline(TestFileOut, TestLineOut);
            write(TestLineOut, DataOut, left, conv_integer(DataOutLength));
            writeline(TestFileOut, TestLineOut);
        end loop;
            DataInLength <= "00001"; -- Crude termination
            Termination <= '1';
            wait for 5 ns;
            StartBypass <= '1';
            wait for 10 ns;
            StartBypass <= '0';
            wait until Finished = '1';
            --write(TestLineOut, conv_integer(DataOutLength), left, OutputWidth);
            --writeline(TestFileOut, TestLineOut);
            write(TestLineOut, DataOut, left, conv_integer(DataOutLength));
            writeline(TestFileOut, TestLineOut);
            
        file_close(TestFileOut);

        wait;
    end process;
	
end struct;

