﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace HEVC_CABAC_Verification_Tool
{
    class CABAC_decoder
    {
        private List<bin> bins_ = new List<bin>();
        private int bins_index;
        private int syntax_element_index;


        private char[] Decoded; 
        private int Decodedindex;

        public RichTextBox binValTarget;
        public RichTextBox debugg;
        uint binValIndex;

        public uint offset;

        static uint qCodIRangeidx, CodIrangeLPS, codIRange, codIOffset, codILow;
        static uint bitsOutstanding;
        static bool firstBitFlag;


        static uint ctxIdxTable_depth = 120;

        public uint[,] rangeTabLPS = new uint[64, 4];
        public uint[] transIdxMPS = new uint[64];
        public uint[] transIdxLPS = new uint[64];
        public uint[] pStateIdxTable = new uint[ctxIdxTable_depth * 3 * 52];
        public uint[] MPSIdxTable = new uint[ctxIdxTable_depth * 3 * 52];


        private void ResetCodeVals()
        {
            try
            {
                //The status of the arithmetic decoding engine is represented by the variables codIRange and codIOffset.
                //In the initialization procedure of the arithmetic decoding process, 
                //codIRange is set equal to 510 and codIOffset is set equal to the value returned from read_bits(9) 
                //interpreted as a 9 bit binary representation of an unsigned integer with most significant bit written first.
                codIRange = 510;
                codILow = 0;
                qCodIRangeidx = 0;
                CodIrangeLPS = 0;
                codIOffset = 0;
                codILow = 0;

                char[] tempA = "XXXXXXXXX".ToCharArray();
                for (int i = 0; i < 9; i++)
                {
                    tempA[i] = (read_bit() == 1) ? '1' : '0';
                }
                codIOffset = Convert.ToUInt32(new string(tempA), 2); // read 9 first binary vals
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }

        }

        public string Decode(List<bin> bins)
        {
            Decoded = new char[65535];
            Decodedindex = 0;

            bins_ = bins;
            bins_index = 0;
            syntax_element_index = 0;

            ResetCodeVals();

            try
            {
                return new string(Decoded).Substring(0, Decodedindex);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }

            return "empty";
        }


        public uint read_bit()
        {
            try
            {
                while (bins_index < bins_.Count)
                {                    
                    return (bins_[bins_index++ ].val == '1') ? (uint)1 : 0;
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }
            MessageBox.Show("read bit called when finished");
            return 0;
        }



        private void DecodeDecicison()
        {
            try
            {
                qCodIRangeidx = (codIRange >> 6) & 3;
                CodIrangeLPS = rangeTabLPS[pStateIdxTable[offset], qCodIRangeidx];
                codIRange = codIRange - CodIrangeLPS;
                if (codIOffset >= codIRange)
                {
                    Decoded[Decodedindex++] = (MPSIdxTable[offset] == 1) ? '0' : '1';
                    codIOffset = codIOffset - codIRange;
                    codIRange = CodIrangeLPS;
                    if (pStateIdxTable[offset] == 0)
                    {
                        MPSIdxTable[offset] = 1 - MPSIdxTable[offset];
                    }
                    pStateIdxTable[offset] = transIdxLPS[pStateIdxTable[offset]];
                }
                else
                {
                    Decoded[Decodedindex++] = ((MPSIdxTable[offset] == 1) ? '1' : '0');
                    pStateIdxTable[offset] = transIdxMPS[pStateIdxTable[offset]];
                }

                RenormD();
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }

        }

        private void DecodeBypass()
        {
            codIOffset = codIOffset << 1;
            codIOffset = codIOffset | read_bit();
            if (codIOffset >= codIRange)
            {
                Decoded[Decodedindex++] = '1';
                codIOffset = codIOffset - codIRange;
            }
            else
            {
                Decoded[Decodedindex++] = '0';
            }
        }

        private void RenormD()
        {
            try
            {
                while (codIRange < 256)
                {
                    codIRange = codIRange << 1;
                    codIOffset = codIOffset << 1;
                    codIOffset = codIOffset | read_bit();
                }
                if (codIOffset >= codIRange)
                {
                    MessageBox.Show("Decoding error:\n The bitstream shall not contain data that result in a value of " +
                                    "codIOffset being greater than or equal to codIRange upon completion of this process.");
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }

        }
    }
}
