﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace HEVC_CABAC_Verification_Tool
{
    class CABAC_encoder
    {
        private List<Syntax_element> S_E_ = new List<Syntax_element>();
        private int S_E_index;
        private List<bin> Encoded;

        public uint offset;

        static uint qCodIRangeidx, CodIrangeLPS, codIRange, codIOffset, codILow;
        static uint bitsOutstanding;
        public bool firstBitFlag;


        static uint ctxIdxTable_depth = 120;

        public uint[,] rangeTabLPS = new uint[64, 4];
        public uint[] transIdxMPS = new uint[64];
        public uint[] transIdxLPS = new uint[64];
        public uint[] pStateIdxTable = new uint[ctxIdxTable_depth * 3 * 52];
        public uint[] MPSIdxTable = new uint[ctxIdxTable_depth * 3 * 52];
        
        public List<bin> Encode(List<Syntax_element> S_E)
        {
            S_E_ = S_E;
            Encoded = new List<bin>();
            bitsOutstanding = 0;
            S_E_index = 0;

            uint bin;
            bool bypass;

            ResetCodeVals();        

            try
            {
                while (S_E_index < S_E.Count)
                {
                    bin = read_bit(out bypass);
                    if (bypass)
                    {
                        EncodeBypass(bin);
                    }
                    else
                    {
                        EncodeDecision(bin);
                    }
                }
                EncodeTerminate(1);

                return Encoded;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }

            return new List<bin>();
        }

        private void init_tables(uint SliceQPY, uint initType)
        {

        }


        public uint read_bit(out bool bypass)
        {
            try
            {
                while (S_E_index < S_E_.Count)
                {
                    if (S_E_[S_E_index].currPos < S_E_[S_E_index].Bins.Count)
                    {
                        bypass = S_E_[S_E_index].bypass;
                        return (S_E_[S_E_index].Bins[S_E_[S_E_index].currPos++] == '1') ? (uint)1 : 0;
                    }
                    else
                    {
                        S_E_index++;
                    }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }
            //MessageBox.Show("read bit called when finished");
            bypass = false;
            return 0;
        }

        public void ResetCodeVals()
        {
            //The status of the arithmetic decoding engine is represented by the variables codIRange and codIOffset.
            //In the initialization procedure of the arithmetic decoding process, 
            //codIRange is set equal to 510 and codIOffset is set equal to the value returned from read_bits(9) 
            //interpreted as a 9 bit binary representation of an unsigned integer with most significant bit written first.
            codIRange = 510;
            codILow = 0;
            qCodIRangeidx = 0;
            CodIrangeLPS = 0;
            codIOffset = 0;
            codILow = 0;
        }
       

        public void EncodeDecision(uint bin)
        {
            qCodIRangeidx = (codIRange >> 6) & 3;
            CodIrangeLPS = rangeTabLPS[pStateIdxTable[offset], qCodIRangeidx];
            codIRange = codIRange - CodIrangeLPS;
            if (bin != MPSIdxTable[offset])
            {
                codILow = codILow + codIRange;
                codIRange = CodIrangeLPS;
                if (pStateIdxTable[offset] == 0)
                {
                    MPSIdxTable[offset] = 1 - MPSIdxTable[offset];
                }
                pStateIdxTable[offset] = transIdxLPS[pStateIdxTable[offset]];
            }
            else
            {
                pStateIdxTable[offset] = transIdxMPS[pStateIdxTable[offset]];
            }
            RenormE();
        }

        public void EncodeBypass(uint bin)
        {
            try
            {
                codILow = codILow << 1;
                if (bin != 0)
                {
                    codILow = codILow + codIRange;
                }
                if (codILow >= 1024)
                {
                    PutBit(1);
                    codILow = codILow - 1024;
                }
                else if (codILow < 512)
                {
                    PutBit(0);
                }
                else
                {
                    codILow = codILow - 512;
                    bitsOutstanding++;
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }
        }

        public void RenormE()
        {
            while (codIRange < 256)
            {
                if (codILow < 256)
                {
                    PutBit(0);
                }
                else if (codILow >= 512)
                {
                    codILow = codILow - 512;
                    PutBit(1);
                }
                else
                {
                    codILow = codILow - 256;
                    bitsOutstanding = bitsOutstanding + 1;
                }
                codIRange = codIRange << 1;
                codILow = codILow << 1;
            }
        }

        //string teststring;

        public void PutBit(uint B)
        {
            if (B != 0 && B != 1) { MessageBox.Show("ERROR: PutBit called with argument: " + B.ToString() + "\n Only 0 or 1 is valid arguments"); }

            if (firstBitFlag)
            {
                firstBitFlag = false;
            }
            else
            {
                bin tempBin = new bin();
                tempBin.val = (B == 1) ? '1' : '0';
                Encoded.Add(tempBin);
            }
            while (bitsOutstanding > 0)
            {
                bin tempBin = new bin();
                tempBin.val = (B == 1) ? '0' : '1';
                Encoded.Add(tempBin);
                bitsOutstanding--;
            }
        }

        public void EncodeTerminate(uint bin)
        {
            try
            {
                codIRange = codIRange - 2;
                if (bin != 0)
                {
                    codILow = codILow + codIRange;
                    EncodeFlush();
                }
                else
                {
                    RenormE();
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }
        }

        public void EncodeFlush()
        {
            try
            {
                codIRange = 2;
                RenormE();
                PutBit((codILow >> 9) & 1);
                //PutBit(((codILow >> 7) & 3) | 1); // does not work. Using a workarround for now.
                PutBit((codILow >> 8) & 1);
                PutBit(((codILow >> 7) & 1) | 1);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }
        }

    }
}
